#version 430 core
/*
* Copyright 2019 Vienna University of Technology.
* Institut of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	flat ivec4 ids;
	vec4 weights;
	vec4 fragPosLightSpace;
} vert;

out vec4 color;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;
uniform sampler2D texture_diffuse1;
uniform sampler2D texture_specular1;
uniform sampler1D celShadingTexture;
uniform bool hasSpecularTexture;

struct DirectionalLight {
	vec3 color;
	vec3 direction;
	vec4 lightSpaceMatrix;
};

struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
};

uniform DirectionalLight directionalLight;
const int NUM_POINT_LIGHTS = 4;
uniform PointLight pointLights[NUM_POINT_LIGHTS];

uniform vec3 ambientLight;
uniform vec4 outlineColor;
uniform int renderContour; //0 = false, 1 = true
uniform float brightnessFactor;
uniform bool celShadingActive;

uniform sampler2DShadow depthMapDirLight;
uniform samplerCubeShadow depthMapPointLights[NUM_POINT_LIGHTS];
uniform float farPlanePointLight;

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

vec3 phongDiffuse(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	return (diffuseF * diffuseC * max(0, dot(n, l))) * att;
}

vec3 phongSpecular(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	vec3 r = reflect(-l, n);
	return (specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att;
}

//returns 0 if it is in shadow, 1 otherwise
float shadowCalculationDir(sampler2DShadow depthMap, vec4 fragPosLightSpace, vec3 lightDirection) {

	//get coords of depthMap
	vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
	projCoords = projCoords * 0.5 + 0.5;

	//TODO adjust parameters when using final game models
	float cosTheta = dot(vert.normal_world, lightDirection);
	float bias = 0.01*tan(acos(cosTheta));
	bias = clamp(bias, 0.003, 0.001);
	projCoords.z -= bias;
	if (projCoords.z > 1.0) {
		return 1.0;
	}
	float closestDepth = texture(depthMap, projCoords);
	return closestDepth;
}

float shadowCalculationPoint(samplerCubeShadow depthMap, vec3 fragPos, vec3 lightPos) {
	vec3 fragToLight = fragPos - lightPos;
	//float closestDepth = texture(depthMap, fragToLight).r;
	//closestDepth *= farPlanePointLight;
	float currentDepth = length(fragToLight);

	float bias = 0.01;
	vec4 d = vec4(fragToLight, (currentDepth / farPlanePointLight) - bias);
	return texture(depthMap, d);
}

void main() {	
	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(camera_world - vert.position_world);
	
	vec3 texColor = texture(texture_diffuse1, vert.uv).rgb;
	
	float shadowCalcD = shadowCalculationDir(depthMapDirLight, vert.fragPosLightSpace, directionalLight.direction);

	color.a = 1;
	color.rgb = texColor * ambientLight * materialCoefficients.x; //ambient
	color.rgb += shadowCalcD * phongDiffuse(n, -directionalLight.direction, v, directionalLight.color * texColor, materialCoefficients.y, directionalLight.color, materialCoefficients.z, specularAlpha, false, vec3(0)); //diffuse directionalL
	for(int i = 0; i < NUM_POINT_LIGHTS; i++) {
		float shadowCalcP = shadowCalculationPoint(depthMapPointLights[i], vert.position_world, pointLights[i].position);
		color.rgb += shadowCalcP * phongDiffuse(n, pointLights[i].position - vert.position_world, v, pointLights[i].color * texColor, materialCoefficients.y, pointLights[i].color, materialCoefficients.z, specularAlpha, true, pointLights[i].attenuation); //diffuse pointL
	}
	

	//cell shading
	if (celShadingActive) {
		color.r = texture(celShadingTexture, color.r).r; //the value of is always in the red channel
		color.g = texture(celShadingTexture, color.g).r;
		color.b = texture(celShadingTexture, color.b).r;
	}
	

	float specularModifier = 0.0;
	if (hasSpecularTexture) {
		specularModifier = texture(texture_specular1, vert.uv).r;
	} else {
		specularModifier = materialCoefficients.z;
	}
	// add directional light specular contribution
	color.rgb += shadowCalcD * phongSpecular(n, -directionalLight.direction, v, directionalLight.color * texColor, materialCoefficients.y, directionalLight.color, specularModifier, specularAlpha, false, vec3(0));
			
	// add point light specular contribution
	for(int i = 0; i < NUM_POINT_LIGHTS; i++) {
		float shadowCalcP = shadowCalculationPoint(depthMapPointLights[i], vert.position_world, pointLights[i].position);
		color.rgb += shadowCalcP * phongSpecular(n, pointLights[i].position - vert.position_world, v, pointLights[i].color * texColor, materialCoefficients.y, pointLights[i].color, specularModifier, specularAlpha, true, pointLights[i].attenuation); //diffuse pointL
	}
	color.rgb *= brightnessFactor;
	
}

